#!/usr/bin/perl
use strict;
use warnings;
use Time::HiRes qw/gettimeofday tv_interval usleep/;

##--( usage )---------------------------------------------------------------------------##
sub usage {
    print STDERR <<"EOH";
qnd_cat_osync.pl v. 0.05 2007/04/13 W.Pilorz
  cat with output synchronization and input buffering
Usage: qnd_cat_osync.pl [--stats] [--cstats] [--obfsz=output_buffer_size] file ... > destfile
  Data from input files is gather into a buffer of size output_buffer_size;
  Output buffer size must be a multiple of 4K;
  Once the buffer is filled up it is written to destination, which is the synced (dup/close).
  After each sync (i.e. dup/close) can sleep for a number of milliseconds specified in
  --sleepms=number_if_milliseconds
  Default obf is 4M.
  -- can be used to indicate end of switches (allows file names starting from dash
EOH
}
##---------------------------------------------------------------------------( usage )--##

my $obf; # output buffer
my $obf_deflt_size = 4 * 1024 * 1024; # default size 4M - as in usage
my $obf_size = undef;
my $stat_req = 0;  # true if --stats or --cstats
my $cstat_req = 0; # true if --cstats
my $sleepms = 0; # numer of milliseconds to sleep after each sync

##--( parse_ARGV )----------------------------------------------------------------------##
sub parse_ARGV {
    my @argv;
    my $had_dashdash = 0; # true if "--" parameter met
    for (@ARGV) {
        if ($had_dashdash) { push @argv, $_; next; }
        if (/^--$/) { $had_dashdash = 1; next; }
        if (/^--stats$/) {  $stat_req = 1; next; }
        if (/^--cstats$/) { $stat_req = 1; $cstat_req = 1; next; }
        if (/^--?h(elp)?$/) { usage(); exit(4); }
        if (/^--sleepms[=:](\d+\.?\d*|\.\d+)/) { $sleepms = $1; next; }
        if (/^--obu?ff?si?ze?[=:](\d+)([kKmM]?)$/) {
            my $sz = $1; my $suf = uc($2);
            $sz *= 1024 if $suf eq "K";
            $sz *= 1024 * 1024 if $suf eq "M";
            die sprintf("ERR: Invalid obfsz value %.0f out of range 4K .. 1024M\n", $sz)
                if $sz < 4*1024 or $sz > 1024*1024*1024;
            die sprintf("ERR: Invalid obfsz value %.0f, not a multiple of 4K \n", $sz)
                if $sz % (4*1024);
            $obf_size = $sz;
            next;
        }
        die "ERR: Unexpected parameter $_, --help for help\n" if /^-/;
        push @argv, $_;
    }
    $obf_size = $obf_deflt_size if not defined $obf_size;
    @ARGV = @argv;
}
##----------------------------------------------------------------------( parse_ARGV )--##

##--( do_sleep )------------------------------------------------------------------------##
sub do_sleep {
    return 0 if not $sleepms;
    my $sv = $sleepms * 1000; # convert to microseconds
    my $seccnt = 100;
    while ($sv > 0.001 and --$seccnt >= 0) {
        my $ret = usleep($sv);
        $sv -= $ret if $ret > 0;
    }
}
##------------------------------------------------------------------------( do_sleep )--##

parse_ARGV();

my $tmcurrent;
sub print_stat {
    my ($pref, $tot, $ptot, $tmstart) = @_;
    my $tm = [ gettimeofday() ];
    my $delta = $tot - $ptot;
    my ($str_tot, $str_delta);
    for ([ \$str_tot, $tot ], [ \$str_delta, $delta ]) {
        my ($rstr, $val) = @$_;
        my $vwrk = $val; my $rem; my @str;
        $rem = $vwrk % 1024; unshift(@str, sprintf("%4d B", $rem)) if $rem;   $vwrk = ($vwrk-$rem)/1024;
        $rem = $vwrk % 1024; unshift(@str, sprintf("%4d KiB", $rem)) if $rem; $vwrk = ($vwrk-$rem)/1024;
                             unshift(@str, sprintf("%4d MiB", $vwrk)) if $vwrk;
        push @str, "0 B" if not @str;
        $$rstr = join(" + ", @str);
    }
    my $tmd = tv_interval($tmstart, $tm); $tmcurrent = $tm;
    my $tmdstr = sprintf "%6.3f s", $tmd;
    my $mibs = "???";
    $mibs = sprintf "%6.3f MiB/s", $delta / 1024 / 1024 / $tmd if $tmd > 0.001;
    print STDERR "$pref$str_delta, time $tmdstr, speed $mibs";
    print STDERR "; total $str_tot" if $tot != $delta;
    print STDERR "\n";
}

local $/ = \4096; # 4K = 4 * 1024

my $tot = 0; my $ptot = 0;
my $tmstart = [ gettimeofday() ]; $tmcurrent =  [ @$tmstart ];
binmode(STDOUT);
my $pARGV = "";
while(<>) {
    if ($ARGV ne $pARGV) { binmode(ARGV); $pARGV = $ARGV; }
    $obf .= $_;
    if (length $obf >= $obf_size) {
        print STDOUT $obf;
        $tot += length $obf; $obf = "";
        open(DUP, ">&1") and close(DUP); do_sleep();
        if ($cstat_req) { print_stat("CSTATS: ", $tot, $ptot, $tmcurrent); $ptot = $tot; }
    }
}
if (defined $obf and length $obf) {
    print STDOUT $obf;
    $tot += length $obf; $obf = "";
    open(DUP, ">&1") and close(DUP);
    if ($cstat_req) { print_stat("CSTATS: ", $tot, $ptot, $tmcurrent); $ptot = $tot; }
}
close(STDOUT) or die "ERR: Could not close STDOUT: $!";
if ($stat_req) { print_stat("FINAL STATS: ", $tot, 0, $tmstart); }

exit(0);
# vim:et ts=4 sw=4 smartindent
